@@ -14,6 +14,7 @@
 package main
 
 import (
+	"bytes"
 	"context"
 	"flag"
 	"fmt"
@@ -31,6 +32,7 @@ import (
 
 	"golang.org/x/oauth2"
 	"golang.org/x/oauth2/endpoints"
+	"rsc.io/qr"
 )
 
 // configByHost lists default config for several public hosts.
@@ -46,6 +48,11 @@ var configByHost = map[string]oauth2.Config{
 		ClientSecret: "2b746eea028711749c5062b9fe626fed78d03cc0",
 		Endpoint:     endpoints.GitHub,
 		Scopes:       []string{"repo", "gist", "workflow"}},
+	"gist.github.com": {
+		ClientID:     "b895675a4e2cf54d5c6c",
+		ClientSecret: "2b746eea028711749c5062b9fe626fed78d03cc0",
+		Endpoint:     endpoints.GitHub,
+		Scopes:       []string{"repo", "gist", "workflow"}},
 	// https://gitlab.com/oauth/applications/232663
 	"gitlab.com": {
 		ClientID: "10bfbbf46e5b760b55ce772a262d7a0205eacc417816eb84d37d0fb02c89bb97",
@@ -131,13 +138,11 @@ func printVersion(w io.Writer) {
 
 func parse(input string) map[string]string {
 	lines := strings.Split(string(input), "\n")
-	pairs := map[string]string{}
+	pairs := make(map[string]string, len(lines))
 	for _, line := range lines {
-		parts := strings.SplitN(line, "=", 2)
-		if len(parts) >= 2 {
-			key, value := parts[0], parts[1]
+		if key, value, ok := strings.Cut(line, "="); ok {
 			_, exists := pairs[key]
-			if strings.HasSuffix(key, "[]") && exists {
+			if exists && strings.HasSuffix(key, "[]") {
 				pairs[key] += "\n" + value
 			} else {
 				pairs[key] = value
@@ -198,7 +203,8 @@ func main() {
 		}
 		host := pairs["host"]
 		looksLikeGitLab := strings.HasPrefix(host, "gitlab.") || strings.Contains(pairs["wwwauth[]"], `realm="GitLab"`)
-		looksLikeGitea := strings.Contains(pairs["wwwauth[]"], `realm="Gitea"`)
+		looksLikeGitea := strings.HasPrefix(host, "gitea.") || strings.Contains(pairs["wwwauth[]"], `realm="Gitea"`)
+		looksLikeForgejo := strings.HasPrefix(host, "forgejo.") || strings.Contains(pairs["wwwauth[]"], `realm="Forgejo"`)
 		looksLikeGitHub := strings.HasPrefix(host, "github.") || strings.Contains(pairs["wwwauth[]"], `realm="GitHub"`)
 		urll := fmt.Sprintf("%s://%s", pairs["protocol"], host)
 		c, found := configByHost[host]
@@ -218,7 +224,7 @@ func main() {
 			}
 			c.Scopes = configByHost["gitlab.com"].Scopes
 		}
-		if !found && looksLikeGitea {
+		if !found && (looksLikeGitea || looksLikeForgejo) {
 			c.ClientID = "a4792ccc-144e-407e-86c9-5e7d8d9c3269"
 			c.Endpoint = oauth2.Endpoint{
 				AuthURL:  fmt.Sprintf("%s/login/oauth/authorize", urll),
@@ -301,9 +307,15 @@ func main() {
 		}
 
 		var authURLSuffix string
-		if looksLikeGitHub && pairs["username"] != "" && pairs["username"] != "oauth2" {
-			// https://docs.github.com/en/apps/oauth-apps/building-oauth-apps/authorizing-oauth-apps#1-request-a-users-github-identity
-			authURLSuffix = fmt.Sprintf("&login=%s", pairs["username"])
+		if pairs["username"] != "" && pairs["username"] != "oauth2" {
+			if looksLikeGitHub {
+				// https://docs.github.com/en/apps/oauth-apps/building-oauth-apps/authorizing-oauth-apps#1-request-a-users-github-identity
+				authURLSuffix = fmt.Sprintf("&login=%s", pairs["username"])
+			}
+			if strings.HasSuffix(host, ".googlesource.com") {
+				// https://developers.google.com/identity/protocols/oauth2/web-server#creatingclient
+				authURLSuffix = fmt.Sprintf("&login_hint=%s", pairs["username"])
+			}
 		}
 
 		if token == nil {
@@ -322,7 +334,7 @@ func main() {
 		}
 		// "A capability[] directive must precede any value depending on it and these directives should be the first item announced in the protocol." https://git-scm.com/docs/git-credential
 		fmt.Println("capability[]=authtype")
-		output := map[string]string{}
+		output := make(map[string]string, 5)
 		hostSupportsBearer := host == "bitbucket.org" || host == "codeberg.org" || host == "gitea.com" || looksLikeGitea || strings.HasSuffix(host, ".googlesource.com")
 		authtypeCapable := strings.Contains(pairs["capability[]"], "authtype")
 		if bearer && hostSupportsBearer && authtypeCapable {
@@ -369,11 +381,17 @@ func main() {
 				// six hours
 				storage = "cache --timeout 21600"
 			}
+			helper := "oauth"
+			if device {
+				helper += " -device"
+			}
 			commands = []*exec.Cmd{exec.Command(gitPath, "config", "--global", "--unset-all", "credential.helper"),
 				exec.Command(gitPath, "config", "--global", "--add", "credential.helper", storage),
-				exec.Command(gitPath, "config", "--global", "--add", "credential.helper", "oauth")}
+				exec.Command(gitPath, "config", "--global", "--add", "credential.helper", helper),
+			}
 		} else if args[0] == "unconfigure" {
-			commands = []*exec.Cmd{exec.Command(gitPath, "config", "--global", "--unset-all", "credential.helper", "oauth")}
+			commands = []*exec.Cmd{
+				exec.Command(gitPath, "config", "--global", "--unset-all", "credential.helper", "oauth*")}
 		}
 		for _, cmd := range commands {
 			cmd.Stderr = os.Stderr
@@ -397,7 +415,7 @@ func main() {
 	}
 }
 
-var template string = `<!DOCTYPE html>
+const template string = `<!DOCTYPE html>
 <html lang="en">
 <head>
 	<title>Git authentication</title>
@@ -488,15 +506,54 @@ func getToken(ctx context.Context, c oauth2.Config, authURLSuffix string) (*oaut
 func getDeviceToken(ctx context.Context, c oauth2.Config) (*oauth2.Token, error) {
 	deviceAuth, err := c.DeviceAuth(ctx)
 	if err != nil {
-		log.Fatalln(err)
+		return nil, err
 	}
 	if verbose {
 		fmt.Fprintf(os.Stderr, "%+v\n", deviceAuth)
 	}
-	fmt.Fprintf(os.Stderr, "Please enter code %s at %s\n", deviceAuth.UserCode, deviceAuth.VerificationURI)
+	if deviceAuth.VerificationURIComplete != "" {
+		fmt.Fprintf(os.Stderr, "Please scan the QR code or enter code %s at %s\n", deviceAuth.UserCode, deviceAuth.VerificationURI)
+		fmt.Fprintln(os.Stderr)
+		writeQRCode(os.Stderr, deviceAuth.VerificationURIComplete)
+	} else {
+		fmt.Fprintf(os.Stderr, "Please enter code %s at %s\n", deviceAuth.UserCode, deviceAuth.VerificationURI)
+	}
 	return c.DeviceAccessToken(ctx, deviceAuth)
 }
 
+func writeQRCode(w io.Writer, data string) error {
+	// use low redundancy to generate small QR codes for terminal output.
+	// we assume the user is sitting in front of their screen so image quality shouldn't be an issue
+	code, err := qr.Encode(data, qr.L)
+	if err != nil {
+		return err
+	}
+
+	var (
+		black     = "\033[40m  \033[0m"
+		white     = "\033[107m  \033[0m"
+		whiteLine = strings.Repeat(white, code.Size+2) + "\n"
+		buf       bytes.Buffer
+	)
+	buf.WriteString(whiteLine)
+	for y := range code.Size {
+		buf.WriteString(white)
+		for x := range code.Size {
+			if code.Black(x, y) {
+				buf.WriteString(black)
+			} else {
+				buf.WriteString(white)
+			}
+		}
+		buf.WriteString(white)
+		buf.WriteRune('\n')
+	}
+	buf.WriteString(whiteLine)
+
+	_, err = buf.WriteTo(w)
+	return err
+}
+
 func replaceHost(e oauth2.Endpoint, host string) oauth2.Endpoint {
 	e.AuthURL = replaceHostInURL(e.AuthURL, host)
 	e.TokenURL = replaceHostInURL(e.TokenURL, host)
